"""Entry repository commands."""

from typing import Literal

import pydantic

from karp_lex_types.value_objects import UniqueId, make_unique_id

from .base import Command


class CreateEntryRepository(Command):
    """Command for creating EntryRepository."""

    id: UniqueId = pydantic.Field(default_factory=make_unique_id)
    name: str
    connection_str: str | None = None
    config: dict
    message: str
    user: str
    cmdtype: Literal["create_entry_repository"] = "create_entry_repository"

    @classmethod
    def from_dict(  # noqa: D102
        cls, data: dict, *, user: str, message: str | None = None
    ) -> "CreateEntryRepository":
        return cls(
            name=data.pop("resource_id"),
            connectionStr=data.pop("connection_str", None),
            config=data,
            user=user,
            message=message or "Entry repository created",
        )

    @pydantic.field_serializer("id")
    def serialize_id(self, id: UniqueId, _info) -> str:  # noqa: PLR6301, A002
        """Serialize id as string."""
        return str(id)



from datetime import datetime  # noqa: D100
from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict, model_validator

from karp_lex_types import alias_generators
from karp_lex_types.value_objects import UniqueIdStr

T = TypeVar("T")


class GenericEntryDto(BaseModel, Generic[T]):  # noqa: D101
    entry: T
    last_modified_by: str | None = None
    last_modified: datetime | None = None
    id: UniqueIdStr | None = None
    version: int | None = None
    resource: str | None = None
    message: str | None = None
    discarded: bool = False
    model_config = ConfigDict(alias_generator=alias_generators.to_lower_camel)

    @model_validator(mode="before")
    @classmethod
    def allow_snake_case(cls, values):  # noqa: D102
        if "last_modified" in values:
            values["lastModified"] = values.pop("last_modified")
        if "last_modified_by" in values:
            values["lastModifiedBy"] = values.pop("last_modified_by")
        if "entity_id" in values:
            values["entityId"] = values.pop("entity_id")
        return values

    def serialize(self):  # noqa: D102
        return self.model_dump(by_alias=True, exclude_none=True)


class EntryDto(GenericEntryDto[dict]):  # noqa: D101
    ...



from karp_lex_types.commands import AddEntry, DeleteEntry, EntryCommand, UpdateEntry
from karp_lex_types.value_objects import make_unique_id


class TestDeserializeEntryCommand:
    def test_can_deserialize_add_entry(self) -> None:  # noqa: PLR6301
        data = {
            "cmd": {
                "cmdtype": "add_entry",
                "resourceId": "resource_a",
                "entry": {"baseform": "sko"},
                "message": "add sko",
                "user": "alice@example.com",
            }
        }

        cmd = EntryCommand(**data)

        assert isinstance(cmd.cmd, AddEntry)

    def test_can_deserialize_delete_entry(self) -> None:  # noqa: PLR6301
        data = {
            "cmd": {
                "cmdtype": "delete_entry",
                "resourceId": "resource_a",
                "id": make_unique_id(),
                "version": 1,
                "message": "add sko",
                "user": "alice@example.com",
            }
        }

        cmd = EntryCommand(**data)

        assert isinstance(cmd.cmd, DeleteEntry)

    def test_can_deserialize_update_entry(self) -> None:  # noqa: PLR6301
        data = {
            "cmd": {
                "cmdtype": "update_entry",
                "resourceId": "resource_a",
                "id": make_unique_id(),
                "entry": {"baseform": "sko"},
                "version": 1,
                "message": "add sko",
                "user": "alice@example.com",
            }
        }

        cmd = EntryCommand(**data)

        assert isinstance(cmd.cmd, UpdateEntry)


class TestSerializeEntryCommand:
    def test_can_serialize_add_entry(self) -> None:  # noqa: PLR6301
        data = {
            "cmd": {
                "cmdtype": "add_entry",
                "resourceId": "resource_a",
                "entry": {"baseform": "sko"},
                "message": "add sko",
                "user": "alice@example.com",
            }
        }

        cmd = EntryCommand(**data)

        assert isinstance(cmd.cmd, AddEntry)
        cmd.model_dump_json()

    def test_can_serialize_delete_entry(self) -> None:  # noqa: PLR6301
        data = {
            "cmd": {
                "cmdtype": "delete_entry",
                "resourceId": "resource_a",
                "id": make_unique_id(),
                "version": 1,
                "message": "add sko",
                "user": "alice@example.com",
            }
        }

        cmd = EntryCommand(**data)

        assert isinstance(cmd.cmd, DeleteEntry)
        cmd.model_dump_json()

    def test_can_serialize_update_entry(self) -> None:  # noqa: PLR6301
        data = {
            "cmd": {
                "cmdtype": "update_entry",
                "resourceId": "resource_a",
                "id": make_unique_id(),
                "entry": {"baseform": "sko"},
                "version": 1,
                "message": "add sko",
                "user": "alice@example.com",
            }
        }

        cmd = EntryCommand(**data)

        assert isinstance(cmd.cmd, UpdateEntry)
        cmd.model_dump_json()



"""Handle of unique ids."""

import typing

import ulid
import ulid.codec
from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic_core import CoreSchema, core_schema

# UniqueId = ulid.ULID

UniqueIdPrimitive = ulid.api.api.ULIDPrimitive
# UniqueIdPrimitive = typing.Union[ulid.api.api.ULIDPrimitive, UniqueIdStr]


class UniqueId(ulid.ULID):  # noqa: D101
    @classmethod
    def __get_pydantic_json_schema__(  # noqa: D105, PLW3201
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, typing.Any]:
        json_schema = super().__get_pydantic_json_schema__(core_schema, handler)  # type: ignore [misc]
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema.update(examples=["01BJQMF54D093DXEAWZ6JYRPAQ"])
        return json_schema

    @classmethod
    def __get_pydantic_core_schema__(  # noqa: D105, PLW3201
        cls, source: typing.Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        def _serialize(instance: typing.Any, info: typing.Any) -> typing.Any:
            return str(instance) if info.mode == "json" else instance

        from_any_schema = core_schema.chain_schema(
            [
                core_schema.any_schema(),
                core_schema.no_info_plain_validator_function(cls.validate),
            ]
        )
        return core_schema.json_or_python_schema(
            json_schema=from_any_schema,
            python_schema=core_schema.union_schema(
                [core_schema.is_instance_schema(ulid.ULID), from_any_schema]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                _serialize, info_arg=True
            ),
        )

    @classmethod
    def validate(cls, v) -> "UniqueId":  # noqa: D102
        if isinstance(v, UniqueId):
            return v  # type: ignore
        if isinstance(v, ulid.ULID):
            return v  # type: ignore
        if not isinstance(v, UniqueIdPrimitive):  # type: ignore
            msg = f"Unsupported type ('{type(v)}')"
            raise TypeError(msg)
        try:
            return ulid.parse(v)  # type: ignore
        except ValueError as err:
            msg = "not a valid ULID"
            raise ValueError(msg) from err

    def __repr__(self) -> str:  # noqa: D105
        return f"UniqueId({super().__repr__()})"


UniqueIdType = (ulid.ULID, UniqueId)


def make_unique_id(
    t: ulid.codec.TimestampPrimitive | None = None,
) -> UniqueId:
    """Generate an UniqueId that are sortable.

    >>> from karp_lex_types.value_objects import make_unique_id
    >>> from datetime import datetime
    >>> old_id = make_unique_id(datetime(1999,12,31,23,59,59))
    >>> make_unique_id() > old_id
    True
    """
    return ulid.new() if t is None else ulid.from_timestamp(t)  # type: ignore


parse = ulid.parse

LEN_OF_ULID: int = 26


class UniqueIdStr(str):  # noqa: D101
    @classmethod
    def __get_pydantic_json_schema__(  # noqa: D105, PLW3201
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, typing.Any]:
        json_schema = super().__get_pydantic_json_schema__(core_schema, handler)  # type: ignore [misc]
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema.update(examples=["01BJQMF54D093DXEAWZ6JYRPAQ"])
        return json_schema

    @classmethod
    def __get_pydantic_core_schema__(  # noqa: D105, PLW3201
        cls, source: typing.Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_before_validator_function(
            cls.validate,
            core_schema.str_schema(),
        )

    @classmethod
    def validate(cls, v, _info):  # noqa: D102
        if isinstance(v, UniqueId | ulid.ULID):
            return str(v)
        if not isinstance(v, str):
            raise TypeError("string or UniqueId required")

        if len(v) != LEN_OF_ULID:
            raise ValueError("invalid uniqueid format")

        return cls(v)

    def __repr__(self) -> str:  # noqa: D105
        return f"UniqueIdStr({super().__repr__()})"



"""Entry commands."""

from collections.abc import Iterable
from typing import (
    Annotated,
    Generic,
    Literal,
    TypeVar,
)

import pydantic

from karp_lex_types.value_objects import UniqueId, make_unique_id

from .base import Command

T = TypeVar("T")


class GenericAddEntry(Command, Generic[T]):  # noqa: D101
    id: UniqueId = pydantic.Field(default_factory=make_unique_id)
    resource_id: str
    entry: T
    message: str
    cmdtype: Literal["add_entry"] = "add_entry"

    @pydantic.field_serializer("id")
    def serialize_id(self, id: UniqueId, _info) -> str:  # noqa: PLR6301, A002
        """Serialize id as string."""
        return str(id)


class AddEntry(GenericAddEntry[dict]):
    """Command to add an entry."""


class AddEntries(Command):  # noqa: D101
    resource_id: str
    entries: Iterable[dict]
    message: str
    chunk_size: int = 0


class AddEntriesInChunks(AddEntries):  # noqa: D101
    chunk_size: int


class DeleteEntry(Command):  # noqa: D101
    resource_id: str
    id: UniqueId
    version: int
    message: str | None = None
    cmdtype: Literal["delete_entry"] = "delete_entry"

    @pydantic.field_serializer("id")
    def serialize_id(self, id: UniqueId, _info) -> str:  # noqa: PLR6301, A002
        """Serialize id as string."""
        return str(id)


class ImportEntries(AddEntries):  # noqa: D101
    pass


class ImportEntriesInChunks(AddEntriesInChunks):  # noqa: D101
    pass


class GenericUpdateEntry(Command, Generic[T]):  # noqa: D101
    resource_id: str
    id: UniqueId
    version: int
    entry: T
    message: str
    force: bool = False
    cmdtype: Literal["update_entry"] = "update_entry"

    @pydantic.field_serializer("id")
    def serialize_id(self, id: UniqueId, _info) -> str:  # noqa: PLR6301, A002
        """Serialize id as string."""
        return str(id)


class UpdateEntry(GenericUpdateEntry[dict]):  # noqa: D101
    ...


EntryCommandType = Annotated[
    AddEntry | DeleteEntry | UpdateEntry, pydantic.Field(discriminator="cmdtype")
]


class EntryCommand(pydantic.BaseModel):
    """A common class for Entry commands."""

    cmd: EntryCommandType


class ExecuteBatchOfEntryCommands(pydantic.BaseModel):
    """Represent a batch of `EntryCommand`s."""

    commands: Iterable[EntryCommand]
